import argparse
from ..utils import tsv_string_empty_none
from ..gzip_utils import open_for_write
from ..dto.transcript_coverage import TranscriptCoverage
from ..segmentation import Segmentation
from ..profile_comparison import compare_coverage_streams

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(
            prog = "compare_coverage",
            description = "Coverage profile comparison",
        )
    argparser.add_argument('segmentation', metavar='segmentation.bed', help='Segmentation')
    argparser.add_argument('coverage_control', metavar='control.bedgraph', help='Coverage for control data')
    argparser.add_argument('coverage_experiment', metavar='experiment.bedgraph', help='Coverage for experiment data')
    argparser.add_argument('--prefix', default='', help='Prefix of feature columns (to distinguish samples)')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    segmentation_stream = Segmentation.each_in_file(args.segmentation, header=False)

    control_coverage_profiles = TranscriptCoverage.each_in_file(args.coverage_control, header=False, dtype=int)
    experiment_coverage_profiles = TranscriptCoverage.each_in_file(args.coverage_experiment, header=False, dtype=int)

    with open_for_write(args.output_file) as output_stream:
        feature_names = ['slope', 'slopelog', 'l1_distance', 'polarity_diff']
        prefixed_feature_names = [f'{args.prefix}{name}' for name in feature_names]
        header = ['transcript_id', *prefixed_feature_names]
        print('\t'.join(header), file=output_stream)
        for rec in compare_coverage_streams(segmentation_stream, control_coverage_profiles, experiment_coverage_profiles):
            info = [rec[field] for field in ['transcript_id', *feature_names]]
            print(tsv_string_empty_none(info), file=output_stream)



from sklearn.linear_model import LinearRegression
import numpy as np
from math import log2
from .polarity_score import polarity_score
from .utils import common_subsequence

def comparison_infos(transcript_id, control_profile, experiment_profile, segmentation):
    assert len(control_profile) == len(experiment_profile) == segmentation.segmentation_length
    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)
    info = {
        'transcript_id': transcript_id,
        'slope': slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=False),
        'slopelog': slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=True),
        'l1_distance': l1_distance_by_segment_counts(control_sums, experiment_sums),
        'polarity_diff': polarity_diff(control_profile, experiment_profile),
    }
    return info

def polarity_diff(control_profile, experiment_profile):
    control_polarity = polarity_score(control_profile)
    exp_polarity = polarity_score(experiment_profile)
    if (exp_polarity is None) or (control_polarity is None):
        return None
    return exp_polarity - control_polarity

def segmentation_stops(segmentation):
    stops = [s.start for s in segmentation.segments]
    stops.append(segmentation.segments[-1].stop)
    return stops

def segmentwise_sums(segmentation, profile):
    profile_cumsum = np.hstack([0, np.cumsum(profile)])
    return np.diff(profile_cumsum[segmentation_stops(segmentation)])

    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)


def slope_by_profiles(control_profile, experiment_profile, segmentation, log_mode=False):
    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)
    return slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=log_mode)

def slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=False):
    assert len(segmentation.segments) == len(control_sums) == len(experiment_sums)
    profile_len = segmentation.segmentation_length
    num_segments = segmentation.num_segments

    total_coverage_control = np.sum(control_sums)
    total_coverage_experiment = np.sum(experiment_sums)
    if (total_coverage_control == 0) or (total_coverage_experiment ==0):
        return None

    if log_mode:
        rate_transform = log2
    else:
        rate_transform = lambda x: x

    xs, ys = [], []
    for (segment, control_segment_sum, experiment_segment_sum) in zip(segmentation.segments, control_sums, experiment_sums):
        if (control_segment_sum == 0) and (experiment_segment_sum == 0):
            continue

        # We add pseudocount of 1 to each total coverage of each segment
        mean_control = (control_segment_sum + 1) / (total_coverage_control + num_segments)
        mean_experiment = (experiment_segment_sum + 1) / (total_coverage_experiment + num_segments)
        detrended_profile = mean_experiment / mean_control

        value = rate_transform(detrended_profile)

        coord = (segment.start + segment.stop - 1) / 2
        rel_coord = coord / profile_len

        # mode: 'center'
        xs.append(rel_coord)
        ys.append(value)

    return slope_by_points(xs, ys)

def slope_by_points(xs, ys):
    if len(xs) < 2:
        return None
    xs = np.array(xs)
    model = LinearRegression()
    model.fit(xs.reshape(-1, 1), ys)
    return model.coef_[0]

def l1_distance(control_profile, experiment_profile, segmentation):
    control_sums = segmentwise_sums(segmentation, control_profile)
    experiment_sums = segmentwise_sums(segmentation, experiment_profile)
    assert len(control_sums) == len(experiment_sums) == len(segmentation.segments)
    return l1_distance_by_segment_counts(control_sums, experiment_sums)

def l1_distance_by_segment_counts(control_sums, experiment_sums):
    assert len(control_sums) == len(experiment_sums)
    control_profile_sum = np.sum(control_sums)
    experiment_profile_sum = np.sum(experiment_sums)
    if (control_profile_sum == 0) or (experiment_profile_sum == 0):
        return None
    control_normed = control_sums / control_profile_sum
    experiment_normed = experiment_sums / experiment_profile_sum
    return np.sum(np.abs(control_normed - experiment_normed))

def _coverage_contig_fn(transcript_coverage):
    return transcript_coverage.transcript_id

def _segmentation_contig_fn(segment):
    return segment.chrom

def align_profile_streams(profile_streams):
    '''
    yields tuples: (transcript_id, profiles)
    '''
    keys = [_coverage_contig_fn] * len(profile_streams)
    yield from common_subsequence(profile_streams, key=keys, check_sorted=True)

def align_profile_streams_to_segmentation(segmentation_stream, profile_streams):
    '''
    yields tuples: (transcript_id, (segmentation, *profiles))
    '''
    streams = [segmentation_stream, *profile_streams]
    keys = [_segmentation_contig_fn] + [_coverage_contig_fn] * len(profile_streams)
    yield from common_subsequence(streams, key=keys, check_sorted=True)

def compare_coverage_streams(segmentation_stream, control_coverage_profiles, experiment_coverage_profiles):
    aligned_stream = align_profile_streams_to_segmentation(
        segmentation_stream,
        [control_coverage_profiles, experiment_coverage_profiles]
    )
    for (transcript_id, (segmentation, control_coverage, experiment_coverage)) in aligned_stream:
        yield comparison_infos(transcript_id,
                               control_coverage.coverage,
                               experiment_coverage.coverage,
                               segmentation)



import argparse
import numpy as np
from ..gzip_utils import open_for_write
from ..tsv_reader import each_in_tsv
from .. import utils

def window_around_idx(arr, idx, window_size, drop_none=False):
    if drop_none:
        left_part = utils.drop_none(arr[:idx])
        element = arr[idx]
        right_part = utils.drop_none(arr[idx + 1:])

        if element is not None:
            arr = left_part + [element] + right_part
        else:
            arr = left_part + right_part
        idx = len(left_part)

    half_window_size = window_size // 2
    if idx < half_window_size:
        return arr[0:window_size]
    elif idx - half_window_size + window_size <= len(arr):
        return arr[(idx - half_window_size):(idx - half_window_size + window_size)]
    else:
        return arr[-window_size:]

def standardize_values(values, standardization, window_size, drop_none=True):
    standardized_values = []
    for (idx, val) in enumerate(values):
        window_values = window_around_idx(values, idx, window_size, drop_none=drop_none)
        mean = np.mean(window_values)
        stddev = np.std(window_values)
        if val is not None:
            standardized_value = standardization(val, val_mean=mean, val_stddev=stddev)
            standardized_values.append(standardized_value)
        else:
            standardized_values.append(None)
    return standardized_values

# stddev --> 1
def standardize_stddev(val, val_mean, val_stddev):
    return val_mean + (val - val_mean) / val_stddev

# mean --> 0
def standardize_mean(val, val_mean, val_stddev):
    return val - val_mean

# mean --> 0
# stddev --> 1
# (aka z-score)
def standardize_zscore(val, val_mean, val_stddev):
    return (val - val_mean) / val_stddev

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="adjust_features", description = "Make length-dependend adjustment of features")
    argparser.add_argument('table', metavar='table.tsv', help='Table in tab-separated format')
    argparser.add_argument('--sort-field', dest='sorting_field', required=True, help='Field to sort a table')
    argparser.add_argument('--fields', nargs='*', dest='fields_to_correct', default=[], help='List of fields to correct')
    argparser.add_argument('--prefix', default='adjusted_', help='Prefix for corrected column name')
    argparser.add_argument('--window', metavar='SIZE', dest='window_size', type=int, default=100, help='Size of sliding window (default: %(default)s)')
    argparser.add_argument('--mode', choices=['zero_mean', 'unit_stddev', 'z-score'], default='z-score', help='How to standardize statistics (default: %(default)s)')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.mode == 'zero_mean':
        standardization = standardize_mean
    elif args.mode == 'unit_stddev':
        standardization = standardize_stddev
    elif args.mode == 'z-score':
        standardization = standardize_zscore
    else:
        raise ValueError(f'Unknown mode `{args.mode}`')

    dtype = lambda x: float(x) if x != '' else None
    data = list(each_in_tsv(args.table))

    fields = list(data[0].keys())
    TRANSFORMED_FIELDS_KEY = object() # unique object not to clash with column name
    for row in data:
        row[TRANSFORMED_FIELDS_KEY] = {}
        for field in [args.sorting_field, *args.fields_to_correct]:
            # we don't want to modify original values because type conversion can screw integer values during output
            row[TRANSFORMED_FIELDS_KEY][field] = dtype(row[field])
    data.sort(key=lambda info: dtype(info[args.sorting_field]))

    modified_data = [row.copy() for row in data]
    output_field_names = fields[:]

    for field in args.fields_to_correct:
        output_field_name = f'{args.prefix}{field}'
        output_field_names.append(output_field_name)
        field_values = [dtype(row[field]) for row in data]
        standardized_field_values = standardize_values(field_values, standardization, args.window_size, drop_none=True)
        for idx in range(len(data)):
            modified_data[idx][output_field_name] = standardized_field_values[idx]

    with open_for_write(args.output_file) as output_stream:
        print('\t'.join(output_field_names), file=output_stream)
        for row in modified_data:
            output_values = [row[field] for field in output_field_names]
            print(utils.tsv_string_empty_none(output_values), file=output_stream)



def parse_condition(condition_str):
    k,vs = condition_str.split('=', maxsplit=1)
    return (k, vs.split(','))

# Known issue: filters now treat all attribute values as strings
def create_record_filter(condition_config):
    k,vs = condition_config
    return lambda rec: str(rec.attributes.get(k, '')) in vs



import numpy as np

def polarity_score(coverage):
    coverage = np.array(coverage)
    total_coverage = coverage.sum()
    if total_coverage == 0:
        return None
    normalized_coverage = coverage / total_coverage
    positions = np.linspace(-1, 1, len(coverage))
    return np.dot(normalized_coverage, positions)
